"use strict";

var global = require("./global.js");
var setTimeout = function(fn, time) {
    INLINE_SLICE(args, arguments, 2);
    global.setTimeout(function() {
        fn.apply(void 0, args);
    }, time);
};

//Feature detect set timeout that passes arguments.
//
//Because it cannot be done synchronously
//the setTimeout defaults to shim and later on
//will start using the faster (can be used without creating closures) one
//if available (i.e. not <=IE8)
var pass = {};
global.setTimeout( function(_) {
    if(_ === pass) {
        setTimeout = global.setTimeout;
    }
}, 1, pass);

module.exports = function(Promise, INTERNAL) {
    var util = require("./util.js");
    var ASSERT = require("./assert.js");
    var apiRejection = require("./errors_api_rejection")(Promise);
    var TimeoutError = Promise.TimeoutError;

    var afterTimeout = function Promise$_afterTimeout(promise, message, ms) {
        //Don't waste time concatting strings or creating stack traces
        if (!promise.isPending()) return;
        if (typeof message !== "string") {
            message = TIMEOUT_ERROR + " " + ms + " ms"
        }
        var err = new TimeoutError(message)
        promise._attachExtraTrace(err);
        promise._rejectUnchecked(err);
    };

    var afterDelay = function Promise$_afterDelay(value, promise) {
        promise._fulfill(value);
    };

    Promise.delay = function Promise$Delay(value, ms, caller) {
        if (ms === void 0) {
            ms = value;
            value = void 0;
        }
        if ((ms | 0) !== ms || ms < 0) {
            return apiRejection(POSITIVE_INTEGER_ERROR);
        }
        if (typeof caller !== "function") {
            caller = Promise.delay;
        }
        var maybePromise = Promise._cast(value, caller, void 0);
        var promise = new Promise(INTERNAL);

        if (Promise.is(maybePromise)) {
            if (maybePromise._isBound()) {
                promise._setBoundTo(maybePromise._boundTo);
            }
            if (maybePromise._cancellable()) {
                promise._setCancellable();
                promise._cancellationParent = maybePromise;
            }
            promise._setTrace(caller, maybePromise);
            promise._follow(maybePromise);
            return promise.then(function(value) {
                return Promise.delay(value, ms);
            });
        }
        else {
            promise._setTrace(caller, void 0);
            setTimeout(afterDelay, ms, value, promise);
        }
        return promise;
    };

    Promise.prototype.delay = function Promise$delay(ms) {
        return Promise.delay(this, ms, this.delay);
    };

    Promise.prototype.timeout = function Promise$timeout(ms, message) {
        if ((ms | 0) !== ms || ms < 0) {
            return apiRejection(POSITIVE_INTEGER_ERROR);
        }

        var ret = new Promise(INTERNAL);
        ret._setTrace(this.timeout, this);

        if (this._isBound()) ret._setBoundTo(this._boundTo);
        if (this._cancellable()) {
            ret._setCancellable();
            ret._cancellationParent = this;
        }
        ret._follow(this);
        setTimeout(afterTimeout, ms, ret, message, ms);
        return ret;
    };

};
